import json
import sys
import time
import traceback
import logging.handlers
import os
import argparse
import psycopg2
import re

from datetime import datetime
from multiprocessing import Pool, Manager, Value

import netaddr
from poaupdater import uUtil, uLogging, uBilling, apsapi, openapi

import mypbaapi
import uCorba2
import utils
import loggingUtils
from utils import DB, VpsProperties, IpPool, Task, Host

IDL_DIR = "./idl"

VPS_NORMALIZATION_SCRIPT_NAME = "normalize-vps.py"
HN_WORKING_DIR = "/usr/local/share/oaci"

DEFAULT_LOGGING_LEVEL = logging.DEBUG

DNS_HOSTING = 'DNS Hosting'

# Most time-consuming operation is VPS normalization on node. This script will most of the time just wait 
# for this operation to complete. So it is okay to select pool size larger than CPU count
THREAD_POOL_SIZE = 16

MIGRATION_STAGES = ["normalize_ve",
                    "create_new_subscription",
                    "import_ve",
                    "plesk_post_import",
                    "prepare_disabling_old_subscription",
                    "disable_old_subscription"]
# MIGRATION_STAGES = ["normalize_ve"]

# stop processing subscriptions after this threshold
FAIL_THRESHOLD = 10

fail_counter = None


class Mode:
    NORMAL_RUN = "normal-run"
    DRY_RUN = "dry-run"
    RETRY = "retry-failed-stage"
    SKIP = "skip-failed-stage"


class Status:
    NOT_STARTED = "NOT_STARTED"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"


class Skipped(Exception):
    pass


class App:
    def __init__(self, subscription_id, config_file):
        self.subscription_id = int(subscription_id)
        self.config_file = config_file
        self.openapi = openapi.OpenAPI()
        self.log = loggingUtils.Logger(subscription_id)

        self.log.debug("Getting config from {}".format(config_file))
        with open(config_file) as config_json:
            self.config = json.load(config_json)

        self.log.debug("Getting IM settings")
        self.im_settings = {}
        self.im_settings["url"] = utils.get_paci_property(self.openapi, "c2u.im.server.uri")
        self.im_settings["login"] = utils.get_paci_property(self.openapi, "rest.login")
        self.im_settings["password"] = utils.get_paci_property(self.openapi, "rest.password")

        self.mn_cert = self.config["mn_certificate"]
        self.aps_api = apsapi.API('https://localhost:6308/aps/2/')

        self.log.debug("Get Legacy Subscription")
        self.legacy_subscription = self.openapi.pem.getSubscription(subscription_id=self.subscription_id,
                                                                    get_resources=True)
        self.db = DB()

        im_db_params = {
            'database': utils.get_config_parameter(self.config, "im_db_user", None),
            'user': utils.get_config_parameter(self.config, "im_db_name", None),
            'host': utils.get_config_parameter(self.config, "im_db_ip", None),
            'password': utils.get_config_parameter(self.config, "im_db_pass", None)
        }
        self.im_db = DB(psycopg2.connect(**im_db_params))

        self.log.debug("Get Service Template info")
        self.service_template_id = utils.get_int_config_parameter(self.config, "service_template_id")
        update_migrate_subscription_summary(self.subscription_id, "targetServiceTemplateId", self.service_template_id)

        self.billing_exists = utils.is_billing_exist()
        if self.billing_exists:
            params = {
                'database': self.config["pba_db_user"],
                'user': self.config["pba_db_name"],
                'host': self.config["pba_db_ip"],
                'password': self.config["pba_db_pass"]
            }
            self.pba_db = DB(psycopg2.connect(**params))

            self.log.debug("Get Billing info")
            mypbaapi.init(
                host=self.config["pba_ip"],
                port=5224,
                uri="/RPC2",
                ssl=True,
                username=self.config["pba_user"],
                password=self.config["pba_password"],
                subscription_id=self.subscription_id)

            self.pba_api = mypbaapi.PBAAPI()

            self.plan_id = utils.get_int_config_parameter(self.config, "plan_id")
            self.plan_period_id = utils.get_int_config_parameter(self.config, "plan_period_id")
            update_migrate_subscription_summary(self.subscription_id, "targetPlanId", str(self.plan_id))
            update_migrate_subscription_summary(self.subscription_id, "targetPlanPeriodId", str(self.plan_period_id))

        self.corba = uCorba2.Corba(IDL_DIR)
        corba_port = utils.get_int_config_parameter(self.config, "corba_port", 8354)
        corba_user = utils.get_config_parameter(self.config, "corba_user", "admin")
        corba_password = utils.get_config_parameter(self.config, "corba_password", "1q2w3e")
        self.corba.connect(host=pleskd_properties["orb.endpoint.addresses"], port=corba_port, user=str(corba_user),
                           password=str(corba_password))

    #
    # Migration entry point
    #

    def migrate(self):
        self.prepare_data()

        if args.mode == Mode.DRY_RUN:
            self.dry_run()
        else:
            self.wet_run()

    def prepare_migration_stages(self):
        state_file_name = utils.migration_state_path(self.subscription_id)
        status_file_exists = os.path.isfile(state_file_name)
        if status_file_exists:
            with open(state_file_name) as state_json:
                self.migration_state = json.load(state_json)
            self.log.debug("State read from %s:" % state_file_name)
            self.log.debug(json.dumps(self.migration_state))

            # read properties from previous run
            for variable_name in self.migration_state["properties"]:
                setattr(self, variable_name, self.migration_state["properties"][variable_name])

        else:
            self.log.debug("File with migration state %s not found, starting from beginning." % state_file_name)
            self.migration_state = {
                "properties": {}
            }
            self.migration_state["stages"] = []
            for stage in MIGRATION_STAGES:
                self.migration_state["stages"].append({
                    "name": stage,
                    "status": Status.NOT_STARTED})

    def wet_run(self):
        self.prepare_migration_stages()

        if self.migration_state["stages"][-1]["status"] == Status.COMPLETED:
            self.log.info("Subscription {} already migrated, skipping".format(self.subscription_id))
            raise Skipped()

        failed_stage = next((stage for stage in self.migration_state["stages"] if stage["status"] == Status.FAILED),
                            None)
        if failed_stage:
            if args.mode == Mode.NORMAL_RUN:
                raise Exception("Migration for this subscription has already run and failed on "
                                " phase: '%s'. Examine migration log for details: %s. When the problem is resolved, run migration"
                                " again with '--mode skip-failed-stage' or '--mode retry-failed-stage' parameter in command line, to skip or retry '%s' phase" %
                                (failed_stage["name"], loggingUtils.subscription_log_path(self.subscription_id),
                                 failed_stage["name"]))

        # process all states one by one
        try:
            for stage in self.migration_state["stages"]:
                execute_stage = False
                if stage["status"] == Status.COMPLETED:
                    pass

                elif stage["status"] == Status.NOT_STARTED:
                    execute_stage = True

                elif stage["status"] == Status.FAILED:
                    if args.mode == Mode.RETRY:
                        execute_stage = True
                    elif args.mode == Mode.SKIP:
                        stage["status"] = Status.COMPLETED

                if execute_stage:
                    self.execute_migration_stage(stage)
        finally:
            self.save_migration_state()

    def execute_migration_stage(self, stage):
        self.log.debug("-----------------------------------")
        self.log.info("Executing migration stage: %s" % stage["name"])
        try:
            getattr(self, stage["name"])()
            stage["status"] = Status.COMPLETED
        except:
            stage["status"] = Status.FAILED
            raise

    def dry_run(self):
        self.normalize_ve()  # will be executed in dry-run mode

    def get_persistent_property(self, name):
        return self.migration_state["properties"][name]

    def set_persistent_property(self, name, value):
        self.migration_state["properties"][name] = value

    #
    # Migration phases
    #
    def prepare_data(self):
        self.get_vps_props_from_db()
        self.check_public_ip_pools_in_im()
        self.check_plesk_installed()
        self.check_dns()
        self.get_limits()

    def normalize_ve(self):
        self.log.debug("Copy files to Hardware Node")
        uUtil.runRemoteCmd("mkdir -p %s" % HN_WORKING_DIR, self.hw_node_ip, "root", None)
        uUtil.copyFileToRemote(VPS_NORMALIZATION_SCRIPT_NAME, HN_WORKING_DIR, self.hw_node_ip, "root", None)
        uUtil.copyFileToRemote(self.subscription_limits_file, HN_WORKING_DIR, self.hw_node_ip, "root", None)
        uUtil.copyFileToRemote(self.config_file, HN_WORKING_DIR, self.hw_node_ip, "root", None)
        uUtil.copyFileToRemote(self.vps_params_file, HN_WORKING_DIR, self.hw_node_ip, "root", None)

        normalization_script_parh = "%s/%s" % (HN_WORKING_DIR, VPS_NORMALIZATION_SCRIPT_NAME)
        config_file_path = "%s/%s" % (HN_WORKING_DIR, self.config_file)
        limits_file_path = "%s/%s" % (HN_WORKING_DIR, os.path.basename(self.subscription_limits_file))
        ve_params_file_path = "%s/%s" % (HN_WORKING_DIR, os.path.basename(self.vps_params_file))

        command = "python %s %s %s %s %s %s" % (normalization_script_parh, self.vps_id, config_file_path,
                                                limits_file_path, ve_params_file_path, self.subscription_id)
        if args.mode == Mode.DRY_RUN:
            command += " --dry-run"

        self.log.debug("Executing VPS normalization script on node {}".format(self.hw_node_ip))
        utils.run_remote_cmd(command, self.hw_node_ip, "root", None)

    def create_new_subscription(self):
        ve = self.get_ve_from_node()
        os_uuid = self.find_suitable_os()
        hostname = ve["Hostname"]
        account_id = self.legacy_subscription["owner_id"]

        aps_defaults = """{"aps": { "type": "http://www.parallels.com/infrastructure/virtual-environment/2.1" },"forMigrate": true, "hostname" : "%s", "osId" : "%s"}""" % \
                       (hostname, os_uuid)

        if self.billing_exists:
            new_subscription_id = self.create_subscription_with_billing(account_id, aps_defaults)
        else:
            new_subscription_id = self.create_subscription_without_billing(account_id, aps_defaults)

        update_migrate_subscription_summary(self.subscription_id, "newSubscrId", new_subscription_id)
        self.set_persistent_property("new_subscription_id", new_subscription_id)
        self.log.info("Created new subscription %s" % new_subscription_id)

    def import_ve(self):
        new_subscription_id = self.get_persistent_property("new_subscription_id")
        ve = self.get_ve_from_node()
        vps_id = ve["ID"]

        self.log.debug("Get resources by type 'http://www.parallels.com/paci/vps-hosting/vps/1.2'"
                       .format(new_subscription_id))
        hosting_with_envs = utils.get_request_to_aps(
            "resources?implementing(http://www.parallels.com/paci/vps-hosting/vps/1.2),"
            "eq(subId,%s),select(environments)" % new_subscription_id, self.aps_api, self.mn_cert)

        new_environment_uuid = hosting_with_envs[0]["environments"][0]["aps"]["id"]
        self.set_persistent_property("new_environment_uuid", new_environment_uuid)
        host_name = self.find_hn_in_im_by_ip(self.hw_node_ip)

        headers = {'Content-Type': 'application/json'}
        if self.with_plesk:
            headers['Plesk-Config'] = self.plesk_configuration_aps_id
            if self.with_dns:
                headers['Enable-Dns'] = True

        if self.billing_exists:
            self.adjust_pba_resources(new_subscription_id)

        self.log.info("Starting import VE")
        utils.post_request_to_aps("resources/%s/import/%s/at/%s" % (new_environment_uuid, vps_id, host_name), headers,
                                  self.aps_api, self.mn_cert)
        self.wait_import_task(new_environment_uuid)
        self.log.info("Import task completed")

    def plesk_post_import(self):
        if self.with_plesk:
            new_environment_uuid = self.get_persistent_property("new_environment_uuid")
            # update license validity
            env = utils.get_request_to_aps("resources/%s" % new_environment_uuid, self.aps_api, self.mn_cert)
            try:
                # long synchronous request here, it may fail if network issue happens
                # actually migration is over and 1 time per day this method is called during sync_plesk_version
                # so we consider exception here is not blocking migration process
                utils.get_request_to_aps("resources/%s/runDailyTasks" % env["plesk_panel"]["aps"]["id"], self.aps_api,
                                         self.mn_cert)
            except Exception as error:
                self.log.warn(error)

            panel = utils.get_request_to_aps("resources/%s" % env["plesk_panel"]["aps"]["id"], self.aps_api,
                                             self.mn_cert)
            daily_task_timeout = utils.get_int_config_parameter(self.config, "plesk_daily_task_timeout", 300)

            def check(this):
                global license
                license = utils.get_request_to_aps("resources/%s" % panel["license"]["aps"]["id"], this.aps_api,
                                                   this.mn_cert)
                assert 'isKeyValid' in license

            utils.wait_for(lambda: check(self), delay=10, timeout=daily_task_timeout, retry_exceptions=False)

            license = utils.get_request_to_aps("resources/%s" % panel["license"]["aps"]["id"], self.aps_api,
                                               self.mn_cert)

            cloud = utils.get_request_to_aps("resources?implementing(http://www.parallels.com/aps/plesk/cloud)",
                                             self.aps_api, self.mn_cert)[0]
            if cloud and "enabled" in cloud and cloud["enabled"] and license["isKeyValid"] and self.plesk_key and \
                    self.plesk_key is not 'PLSK.00000000.0000':
                result_code = utils.plesk_ka_get_key_info(cloud["customKAServer"], cloud["kaApiLogin"],
                                                          cloud["kaApiPassword"], self.plesk_key)
                # https://docs.plesk.com/en-US/onyx/partner-api-1.0/specifications-of-methods/partner10getkeyinfo.77570/
                if result_code in ["200", "201", "300", "400", "410"]:
                    self.fatal("Failed to get keyInfo")

                if result_code in ["100"]:
                    utils.put_request_to_aps("resources/%s" % license["aps"]["id"],
                                             {"mainKeyNumber": self.plesk_key,
                                              "retrievedFromBilling": True},
                                             self.aps_api, self.mn_cert)

    def add_leading_zeros(self, ip):
        formatted_map = map(lambda val: '{0:0{width}}'.format(int(val), width=3), ip.split('.'))
        return '.'.join(formatted_map)

    def prepare_disabling_old_subscription(self):
        # set 'soft_remove' flag in DB such that Legacy VPS does not remove this VE during subscribtion cancelation
        self.log.info("Disable old subscription")
        self.db.update("UPDATE ves SET soft_remove='y' WHERE ve_id=%s AND hw_host_id=%s" % (self.vps_id, self.host_id))

        try:
            self.db.update("DELETE FROM psa WHERE host_id = %s" % self.ve_host_id)
        except Exception as e:
            self.log.info("Expected exception during delete from psa: {}".format(str(e)))

        # bind ips to IM to prevent unmarking from poa ip pool
        self.log.info("Bind ips to IM host")
        im_ip = utils.get_config_parameter(self.config, "im_ip", None)
        im_host_id_from_db = self.db.select("select host_id from configured_ips WHERE ip_address = '%s'" %
                                            self.add_leading_zeros(im_ip), lambda row: row[0])
        im_host_id = utils.get_config_parameter(self.config, "im_host_id", im_host_id_from_db[0])
        if_id = self.db.select("select if_id from net_interfaces WHERE host_id = %s" % im_host_id,
                               lambda row: row[0])
        im_if_id = utils.get_config_parameter(self.config, "im_if_id", if_id[0])
        try:
            self.db.update(
                "UPDATE configured_ips set host_id = %s, if_id = %s WHERE host_id=%s" % (im_host_id, im_if_id,
                                                                                         self.ve_host_id))
        except Exception as e:
            # in 17.13 there are 2 same ips in pem pool after migration.
            # one of them which related to legacy will be dropped after host drop
            self.log.info("Expected exception during update configured_ips: {}".format(str(e)))

        if self.with_plesk:
            try:
                self.db.update(
                    "DELETE FROM domain_hosting_bindings WHERE subscription_id = {} AND dhtype = 'Plesk'"
                    .format(self.subscription_id))
            except Exception as e:
                self.log.info("Expected exception during delete domain_hosting_bindings: {}".format(str(e)))

        # Removing host will also remove VPS records from DB. Otherwise, when we cancel old subscribtion,
        # VPS will be disabled on node
        ref = self.corba.reference("Host", "Plesk.Host")
        ref.removeHost(self.ve_host_id)

        # wait until host removed
        while True:
            time.sleep(1)
            self.log.debug("Waiting for removeHost completion for %s" % self.ve_host_id)
            hosts = self.db.select("SELECT host_id FROM hosts WHERE host_id = %s" % self.ve_host_id,
                                   lambda row: Host(row))
            if len(hosts) == 0:
                break

        # mark plesk licenses in KA gate as deleted to prevent terminating them in KA during cancel subscription
        if self.with_plesk:
            self.log.info("Mark plesk licenses in KA gate as marked")
            try:
                self.pba_db.update(
                    'UPDATE "License" SET "Status"=700 WHERE "subscriptionID" = %d' % self.subscription_id)
            except Exception as e:
                self.log.info("Expected exception update \"License\" table, KAGATE is not installed")

    def disable_old_subscription(self):
        new_subscription_id = self.get_persistent_property("new_subscription_id")
        if self.billing_exists:
            # 10 - customer is refunded, 20 - customer account is credited, 30 - customer is not refunded
            subscription_cancelation_type = utils.get_config_parameter(self.config, "subscription_cancelation_type", 30)
            subscription_cancelation_reason_id = utils.get_config_parameter(self.config,
                                                                            "subscription_cancelation_reason_id", 1)
            subscription_cancelation_comment = utils.get_config_parameter(self.config,
                                                                          "subscription_cancelation_comment",
                                                                          "Migrated to subscription #%s"
                                                                          % new_subscription_id)

            self.pba_api.PlaceSubscriptionCancellationOrder_API(self.subscription_id, subscription_cancelation_type,
                                                                subscription_cancelation_reason_id,
                                                                subscription_cancelation_comment)
        else:
            self.openapi.pem.disableSubscription(subscription_id=self.subscription_id)

    #
    # Utility stuff
    #
    def save_migration_state(self):
        with open(utils.migration_state_path(self.subscription_id), 'w') as outfile:
            json.dump(self.migration_state, outfile, indent=4)

    def fatal(self, msg):
        uLogging.err("[SubId: {}] Fatal Exception: {}".format(self.subscription_id, msg))
        raise Exception("ABORTING: %s" % msg)

    def get_os_templates_from_im(self):
        resp = utils.request_to_im("template", self.im_settings)
        return resp["template"]

    def find_hn_in_im_by_ip(self, ip):
        resp = utils.request_to_im("hn", self.im_settings)
        for node in resp["node"]:
            full_node = utils.request_to_im("hn/%s" % node["@name"], self.im_settings)
            for interface in full_node["network-interfaces"]["net-if"]:
                if interface["@ip"] == str(ip):
                    return node["@name"]

    def get_vps_props_from_db(self):
        props = self.db.select("SELECT h.primary_name, h2.host_id, v.ve_id, hn.agent_ip, v.ip_address, v.ve_host_id "
                               "FROM hosts h, hosts h2, ves v, hw_nodes hn "
                               "WHERE h.subscription_id = %s AND h.host_id = v.ve_host_id "
                               "AND h2.host_id = v.hw_host_id AND v.hw_host_id=hn.host_id"
                               % self.subscription_id, lambda row: VpsProperties(row))

        if not props:
            self.fatal("Server at subscription {} not found at DB".format(self.subscription_id))

        self.log.debug("Read from DB: vps name: %s, hardware node name: %s" % (props[0].vps_name, props[0].host_id))
        self.vps_name = props[0].vps_name
        self.host_id = props[0].host_id
        self.vps_id = props[0].vps_id
        host_ip_with_leading_zeroes = props[0].host_ip_with_leading_zeroes
        vps_ip_with_leading_zeroes = props[0].vps_ip_with_leading_zeroes
        self.ve_host_id = props[0].ve_host_id

        self.hw_node_ip = netaddr.IPAddress(host_ip_with_leading_zeroes, flags=netaddr.ZEROFILL).ipv4()

        # find gateway for vps (legacy does not have one because it uses host-routed NIC)
        self.log.debug("Read from DB network settings")
        addr = netaddr.IPAddress(vps_ip_with_leading_zeroes, flags=netaddr.ZEROFILL)

        pools = self.db.select("SELECT ip_from, ip_to, gateway, pool_name FROM ip_pools", lambda row: IpPool(row))
        pool_found = False
        for pool in pools:
            from_ip = netaddr.IPAddress(pool.ip_from, flags=netaddr.ZEROFILL)
            to_ip = netaddr.IPAddress(pool.ip_to, flags=netaddr.ZEROFILL)

            if addr in netaddr.IPRange(from_ip, to_ip):
                pool_found = True
                if not pool.gateway:
                    self.fatal(
                        "IP Pool '%s' does not have a gateway. Assign gateway and run migration again." % pool.pool_name)

                vps_gw = netaddr.IPAddress(pool.gateway, flags=netaddr.ZEROFILL).ipv4()

        if not pool_found:
            self.fatal("Server IP address %s does not belong to any IP pool."
                       "Register IP pool and run migration again" % addr)

        vps_params = {}
        vps_params["gateway"] = str(vps_gw)

        self.vps_params_file = utils.vps_params_path(self.subscription_id)
        with open(self.vps_params_file, 'w') as outfile:
            json.dump(vps_params, outfile, indent=4)

    def check_public_ip_pools_in_im(self):
        self.log.debug("Check if IP`s can be bound to IM pools and those pools are disabled")
        ip_addresses = self.db.select("select ip_address from configured_ips WHERE host_id = %s" % self.ve_host_id,
                                      lambda row: row[0])
        pools = self.im_db.select("select ip_min, ip_max, gateway, label from public_nets where disabled is true",
                                  lambda row: IpPool(row))
        for ve_ip in ip_addresses:
            ve_ip_addr = netaddr.IPAddress(ve_ip, flags=netaddr.ZEROFILL)
            im_pool_found = False
            for pool in pools:
                ip_min = netaddr.IPAddress(pool.ip_from, flags=netaddr.ZEROFILL)
                ip_max = netaddr.IPAddress(pool.ip_to, flags=netaddr.ZEROFILL)
                if ve_ip_addr in netaddr.IPRange(ip_min, ip_max):
                    im_pool_found = True
                    break
            if not im_pool_found:
                self.fatal("Disabled IP pool in IM not found for server IP %s." % ve_ip_addr)

    def get_limits(self):
        if self.billing_exists:
            limits = self.get_limits_for_plan(self.plan_id)
        else:
            limits = self.get_limits_for_st(self.service_template_id)

        self.validate_limits(limits)
        self.subscription_limits_file = utils.subscription_limits_path(self.subscription_id)
        with open(self.subscription_limits_file, 'w') as outfile:
            json.dump(limits, outfile, indent=4)

    def validate_limits(self, limits):
        def report_error(res_name):
            self.fatal(
                "%s resource not found in Service Plan #%d nor in its Service Template" % (res_name, self.plan_id))

        if not ("cpu" in limits):
            report_error("VPS CPU Number")
        if not ("ram" in limits):
            report_error("VPS Memory")
        if not ("hdd" in limits):
            report_error("VPS Diskspace")
        if not ("bandwidth" in limits):
            report_error("VPS Bandwidth")

    def plan_includes_to_pem_limits(self, plan_id):
        limits = []
        resp = self.pba_api.PlanRateListFullGet_API(plan_id, 1)

        for rate in resp:
            resource_id = rate[10]
            included = rate[3]
            uom = rate[6]

            if self.adjust_ip and self.adjust_ip > 0 and resource_id == self.ip_resource:
                limits.append({"resource_id": resource_id, "resource_limit": included + self.adjust_ip})
            elif self.with_plesk and resource_id == self.plesk_configuration:
                limits.append({"resource_id": self.plesk_configuration, "resource_limit": 1})
            else:
                limits.append(
                    {"resource_id": resource_id, "resource_limit": utils.convert_plan_limit_to_pem_uom(included, uom)})
        return limits

    def check_st_contains_plesk_resources(self, service_template, plan_rates=None):
        self.log.debug("Check that Service Template contains Plesk resources")
        panel_exists = False
        license_exists = False

        # check PVPS_Panel
        for resource_type in service_template["resource_types"]:
            for act_param in resource_type["act_params"]:
                if act_param["var_value"].startswith("http://www.parallels.com/aps/plesk/panel/") and \
                        resource_type["limit"] == "1":
                    panel_exists = True
                    break
        if not panel_exists:
            self.fatal(
                "http://www.parallels.com/aps/plesk/panel/ resource is missing in service template or limit != 1")

        # check PVPS_License
        for resource_type in service_template["resource_types"]:
            for act_param in resource_type["act_params"]:
                if act_param["var_value"].startswith("http://www.parallels.com/aps/plesk/license/") and \
                        resource_type["limit"] == "1":
                    license_exists = True
                    break
        if not license_exists:
            self.fatal(
                "http://www.parallels.com/aps/plesk/license/ resource is missing in service template or limit != 1")

        # check configuration
        configurations = self.get_plesk_configurations_from_template(service_template)
        if len(configurations) == 0:
            self.fatal("Plesk configurations are not found within service template")
        if "plesk_configuration" in self.config:
            conf_list = filter(lambda conf: conf["resource_type_id"] == self.config["plesk_configuration"],
                               configurations)
            if len(conf_list) == 1:
                self.plesk_configuration = conf_list[0]["resource_type_id"]
                self.plesk_configuration_aps_id = conf_list[0]["aps_id"]
            else:
                self.fatal("Plesk configuration from migration_params.json not found in service template")
        else:
            if len(configurations) == 1:
                self.plesk_configuration = configurations[0]["resource_type_id"]
                self.plesk_configuration_aps_id = configurations[0]["aps_id"]
            else:
                self.fatal("Multiple plesk configurations are found within service template, "
                           "please define plesk_configuration property in migration_params.json")

        # check that only self.plesk_configuration has limit = 1 in ST
        for conf in configurations:
            if conf["resource_type_id"] != self.plesk_configuration:
                if conf["limit"] == "1":
                    self.fatal("Plesk configuration %s limit must be 0" % conf["resource_type_id"])
            else:
                if conf["limit"] != "1" and (plan_rates is None or
                                             len(filter(lambda rate: rate[10] == conf["resource_type_id"],
                                                        plan_rates)) == 0):
                    if self.billing_exists:
                        self.fatal("Plesk configuration %s limit must be 1 in service template or "
                                   "Plesk configuration must be part of service plan resources" %
                                   conf["resource_type_id"])
                    else:
                        self.fatal("Plesk configuration %s limit must be 1 in service template" %
                                   conf["resource_type_id"])

    def get_limits_for_plan(self, plan_id):
        limits = {}
        plan_limits_by_rt_id = {}

        # fetch limits from plan and put in map by RT ID
        plan_rates = self.pba_api.PlanRateListFullGet_API(plan_id, 1)

        for rate in plan_rates:
            rt_id = rate[10]
            # this is included in fact, these parameters passed to normalize script and check through real VPS
            limit = rate[3]
            plan_limits_by_rt_id[rt_id] = int(limit)

        # fetch resources and limits from ST
        resp = self.pba_api.PlanDetailsGet_API(plan_id)
        st_id = resp[17]
        self.log.debug("Get Service Template")
        st = self.openapi.pem.getServiceTemplate(st_id=st_id, get_resources=True, get_full_info=True)
        self.check_ip_limit(service_template=st, plan_rates=plan_rates)
        if self.with_dns:
            self.check_dns_in_st(service_template=st)

        self.check_plan_not_contains_included_configuration(plan_rates, self.get_plesk_configurations_from_template(st))
        if self.with_plesk:
            self.check_st_contains_plesk_resources(service_template=st, plan_rates=plan_rates)
        for rt in st["resource_types"]:
            if not ("act_params" in rt):
                continue

            if (any(("var_value" in param and param["var_value"] == "vps-hosting") and (
                    "var_name" in param and param["var_name"] == "service_id") for param in rt["act_params"])):
                counter_name = next(
                    (param["var_value"] for param in rt["act_params"] if param["var_name"] == "resource_id"), None)
                resource_type_id = rt["resource_type_id"]

                # if RT is in plan - take limit from plan. If not - from ST.
                if resource_type_id in plan_limits_by_rt_id:
                    self.log.debug("limit for {} found in plan".format(resource_type_id))
                    limit = plan_limits_by_rt_id[resource_type_id]
                    # assume that limits for disk and backup in plan are in GB
                    if counter_name in ["hdd", "backupSpace"]:
                        limit = limit * 1024  # MB

                    limits[counter_name] = limit

                else:
                    self.log.debug("limit for {} NOT found in plan, taking from ST".format(resource_type_id))
                    limit = int(rt["limit"])
                    # convert to ram and hdd from KB (platform) to MB (prlctl)
                    if counter_name in ["ram", "hdd", "backupSpace"]:
                        limit = limit / 1024

                    limits[counter_name] = limit
        return limits

    def get_limits_for_st(self, st_id):
        limits = {}

        self.log.debug("Get Service Template")
        st = self.openapi.pem.getServiceTemplate(st_id=self.service_template_id, get_resources=True, get_full_info=True)
        self.check_ip_limit(service_template=st, plan_rates=None)
        if self.with_dns:
            self.check_dns_in_st(service_template=st)
        if self.with_plesk:
            self.check_st_contains_plesk_resources(service_template=st)
        for rt in st["resource_types"]:
            if not ("act_params" in rt):
                continue

            if (any(("var_value" in param and param["var_value"] == "vps-hosting") and (
                    "var_name" in param and param["var_name"] == "service_id") for param in rt["act_params"])):
                counter_name = next(
                    (param["var_value"] for param in rt["act_params"] if param["var_name"] == "resource_id"), None)
                limit = int(rt["limit"])
                # convert to ram and hdd from KB (platform) to MB (prlctl)
                if counter_name in ["ram", "hdd", "backupSpace"]:
                    limit = limit / 1024

                limits[counter_name] = limit

        return limits

    def get_ve_os(self):
        self.log.debug("Get server's OS Template")
        resp = uUtil.runRemoteCmd("vzpkg list -q %s | awk 'NR==1{print $1}'" % self.vps_id, self.hw_node_ip, "root",
                                  None)
        return resp.strip()

    def find_suitable_os(self):
        self.log.debug("Find suitable OS Template")
        os_name_on_node = self.get_ve_os()
        self.log.debug("OS Template {} found at Node".format(os_name_on_node))
        im_templates = self.get_os_templates_from_im()
        suitable_template_from_im = next((template["@name"] for template in im_templates if (
                ("@name-on-hn" in template) and (template["@name-on-hn"] == os_name_on_node))), None)
        self.log.debug("OS Template {} found at IM".format(suitable_template_from_im))

        if not suitable_template_from_im:
            self.log.warn("OS template {} associated with vps {} is not registered in IM"
                          .format(os_name_on_node, self.vps_name))
            if not ("default_os_template" in self.config):
                self.fatal(
                    "OS Template %s not found, default_os_template not specified in parameters" % os_name_on_node)

            suitable_template_from_im = next((template["@name"] for template in im_templates if (
                    ("@name-on-hn" in template) and (template["@name-on-hn"] == os_name_on_node))), None)
            if suitable_template_from_im:
                self.log.warn("Using default OS Template from config: {}".format(self.config["default_os_template"]))
            else:
                self.fatal("Default OS Template %s specified in configuration not found in IM" % self.config[
                    "default_os_template"])

        aps_templates = utils.get_request_to_aps(
            "resources?implementing(http://www.parallels.com/infrastructure/virtual-environment/os-template/1.0),eq(name,%s)" % suitable_template_from_im,
            self.aps_api, self.mn_cert)

        if not aps_templates:
            self.fatal("OS template %s is not registered in APS."
                       "Make sure it is listed in IM (Cloud Infrastructure -> OS Templates) first."
                       "Synchronization with APS can be performed by 2017-10-30-CCU-17112-synch-os-templates.post.py")

        return aps_templates[0]["aps"]["id"]

    def get_ve_from_node(self):
        self.log.info("Get Server {} from Node {}".format(self.vps_id, self.hw_node_ip))
        ve_json = uUtil.runRemoteCmd("prlctl list -i %s --json" % self.vps_id, self.hw_node_ip, "root", None)
        ves = json.loads(ve_json)
        return ves[0]

    def calculate_start_exp_date(self, period, period_type):
        # subscription period type: 0 - Hour(s), 1 - Day(s), 2 - Month(s), 3 - Year(s);
        if not (period_type in [2, 3]):
            self.fatal("Only months and years are supported in subscription period."
                       "Specified subscription period has type %s. (0 - Hour(s), 1 - Day(s), 2 - Month(s), 3 - Year(s))"
                       % period_type)
            return None

        now = datetime.now()

        months_total = period
        if period_type == 3:
            months_total = period * 12

        exp_month = (now.date().month + months_total) % 12
        exp_year = now.date().year + (now.date().month + months_total) / 12
        exp_day = now.date().day

        self.log.debug("exp_month={}, exp_year={}, exp_day={}".format(exp_month, exp_year, exp_day))

        while True:
            try:
                exp_date = now.replace(month=exp_month, year=exp_year, day=exp_day)
                break
            except ValueError:
                exp_day -= 1

        return utils.datetime_to_timestamp(now), utils.datetime_to_timestamp(exp_date)

    def create_subscription_with_billing(self, account_id, aps_defaults):
        # get plan details
        resp = self.pba_api.PlanDetailsGet_API(self.plan_id)

        billing_period_type = resp[10]
        billing_period = resp[11]
        service_template_id = resp[17]

        # get period details
        resp = self.pba_api.PlanPeriodDetailsGet_API(self.plan_period_id)

        period_duration = resp[2]
        period_type = resp[3]
        setup_fee = resp[5]
        subscription_fee = resp[6]
        renewal_fee = resp[7]
        transfer_fee = resp[8]
        non_refundable_amt = resp[9]
        refund_period = resp[10]
        start_and_exp_date = self.calculate_start_exp_date(period_duration, period_type)

        # get old subscription details
        resp = self.pba_api.SubscriptionDetailsGetEx_API(self.subscription_id)

        last_bill_date = resp[9]
        next_bill_date = resp[10]

        old_subscr_billing_period_type = resp[11]
        old_subscr_billing_period = resp[12]

        if not (billing_period_type == old_subscr_billing_period_type):
            self.fatal(
                "Billing period type for existing subscription is %s but for new plan it is %s. They must be the same" % (
                    old_subscr_billing_period_type, billing_period_type))

        if not (billing_period == old_subscr_billing_period):
            self.fatal(
                "Billing period for existing subscription is %s but for new plan it is %s. They must be the same" % (
                    billing_period, old_subscr_billing_period))

        global pba_lock
        with pba_lock:
            resp = self.pba_api.AddPEMSubscription_API(
                30,  # Status;
                50,  # ServStatus;
                start_and_exp_date[0],  # StartDate;
                start_and_exp_date[1],  # ExpirationDate;
                account_id,  # AccountID;
                self.plan_id,  # PlanID;
                period_duration,  # Period duration;
                period_type,  # PeriodType 0 - Hours, 1 - Days, 2 - Months, 3 - Years;
                "",  # DomainName;
                setup_fee,  # SetupFee;
                subscription_fee,  # SubscriptionFee;
                renewal_fee,  # RenewalFee;
                transfer_fee,  # TransferFee;
                non_refundable_amt,  # NonRefundableAmt;
                refund_period,  # RefundPeriod;
                billing_period_type,  # BillingPeriodType;
                billing_period,  # BillingPeriod;
                last_bill_date,  # LastBillDate;
                next_bill_date  # NextBillDate.
            )

        subscription_id = resp[0]

        params = [{"var_name": "aps_defaults_1", "var_value": aps_defaults}]
        resource_limits = self.plan_includes_to_pem_limits(self.plan_id)
        self.log.debug("Activate new Subscription for account {} and Service Template {}"
                       .format(account_id, service_template_id))
        resp = self.openapi.pem.activateSubscription(account_id=account_id, subscription_id=subscription_id,
                                                     service_template_id=service_template_id, parameters=params,
                                                     resource_limits=resource_limits)

        return resp["subscription_id"]

    def create_subscription_without_billing(self, account_id, aps_defaults):
        params = [{"var_name": "aps_defaults_1", "var_value": aps_defaults}]
        self.log.debug("Activate new Subscription for account {} and Service Template{}"
                       .format(account_id, self.service_template_id))
        res = self.openapi.pem.activateSubscription(account_id=account_id, service_template_id=self.service_template_id,
                                                    parameters=params)
        return res["subscription_id"]

    def increase_plesk_pba(self, subscription_id):
        self.log.debug("Increase Plesk at PBA")
        self.adjust_resource_in_subscription(subscription_id, self.plesk_configuration, 1.0)

    def adjust_resource_in_subscription(self, subscription_id, resource_id, amount):
        self.pba_api.AddResourceToSubscription_API(
            int(subscription_id),  # SubscriptionID;
            int(resource_id),  # ResourceID;
            float(amount),  # Amount;
            2,  # Installed = 2                 # Status;
            "",  # SetupFee;
            "",  # RecurringFee;
            ""  # CostForAdditional;
        )

    def wait_import_task(self, env_uuid):
        check = utils.check_task(self.db, "Import",
                                 "SELECT task_id, exec_status FROM tm_logs "
                                 "WHERE name LIKE '%%%s%%' AND name LIKE '%%%s%%' AND "
                                 "method = 'APSAsyncOperations' ORDER BY finished DESC limit 1;" % ("import", env_uuid))
        self.log.debug("Waiting for task Import will be completed")
        utils.wait_for(check, delay=10, timeout=3600, retry_exceptions=False)

    def check_plesk_installed(self):
        self.with_plesk = False
        for resource_type in self.legacy_subscription["resource_types"]:
            if resource_type["resclass_name"] == "Plesk" and resource_type["usage"] == "1":
                self.with_plesk = True

        self.log.info("self.check_plesk_installed() is %r" % self.with_plesk)

        if self.with_plesk:
            raw_plesk_key = uUtil.runRemoteCmd(
                "vzctl exec %s \"plesk bin keyinfo --list | grep plesk_key_id\"" % self.vps_id,
                self.hw_node_ip, "root", None)
            # plesk_key_id: PLSK.10468469.0001
            m = re.search('plesk_key_id:\s*(.*)', raw_plesk_key)
            if m:
                found = m.group(1)
                self.plesk_key = found[:4] + '.' + found[4:-4] + '.' + found[-4:]
                if len(self.plesk_key) != 18:
                    self.fatal("Invalid plesk_key {}".format(self.plesk_key))
            else:
                self.fatal("Failed to get plesk key {}".format(raw_plesk_key))

            self.log.info("plesk_key is {}".format(self.plesk_key))

    def check_plan_not_contains_included_configuration(self, plan_rates, plesk_configurations):
        for rate in plan_rates:
            rt_id = int(rate[10])
            included = int(rate[3])
            if len(filter(lambda conf: conf["resource_type_id"] == rt_id, plesk_configurations)) != 0 and included != 0:
                self.fatal("Plesk configuration with id=%d is included in service plan" % rt_id)

    def get_plesk_configurations_from_template(self, service_template):
        self.log.info("Get Plesk configurations from template")
        configurations_response = utils.get_request_to_aps(
            "resources?implementing(http://www.parallels.com/aps/plesk/configuration)", self.aps_api, self.mn_cert)
        configurations = {}
        for conf in configurations_response:
            configurations[conf["aps"]["id"]] = conf["details"]["enableKaIntegration"]
        if len(configurations) == 0:
            self.fatal("Plesk configurations are not found")
        result = []
        for resource_type in service_template["resource_types"]:
            for act_param in resource_type["act_params"]:
                if configurations.get(act_param["var_value"]) is not None:
                    result.append({"resource_type_id": resource_type["resource_type_id"],
                                   "limit": resource_type["limit"],
                                   "aps_id": act_param["var_value"]})
        return result

    def check_dns_in_st(self, service_template):
        if not any("resclass_name" in rt and rt["resclass_name"] == DNS_HOSTING
                   for rt in service_template["resource_types"]):
            self.fatal("Could not find \"{}\" resource in service template".format(DNS_HOSTING))

    def check_ip_limit(self, service_template, plan_rates):
        legacy_ip_usage = 0
        for resource_type in self.legacy_subscription["resource_types"]:
            if resource_type["resclass_name"] == "ips":
                legacy_ip_usage = int(resource_type["usage"])
        self.log.info("legacy_ip_usage is %d" % legacy_ip_usage)

        ip_resource_found = False
        for rt in service_template["resource_types"]:
            if not ("act_params" in rt):
                continue

            if (any("var_value" in param and
                    param["var_value"].startswith(
                        "http://www.parallels.com/infrastructure/virtual-environment/ip-v4-address") for param in
                    rt["act_params"])):
                ip_resource_found = True
                # rate[10] - resourceId, rate[4] - limit
                if plan_rates and any(rate[10] == rt["resource_type_id"] for rate in plan_rates):
                    for rate in plan_rates:
                        if rate[10] == rt["resource_type_id"]:
                            if legacy_ip_usage > int(rate[4]):
                                self.fatal(
                                    "Resource limit for IP in service plan ({}) is less than subscriptions usage ({})".
                                        format(int(rate[4]), legacy_ip_usage))
                            else:
                                # amount of IP to upsell, int(rate[3]) is included
                                self.adjust_ip = legacy_ip_usage - int(rate[3])
                                self.ip_resource = int(rate[10])
                else:
                    if legacy_ip_usage > int(rt["limit"]):
                        self.fatal("Service template IP limit ({}) is less than subscriptions usage ({})".
                                   format(int(rt["limit"]), legacy_ip_usage))

        if legacy_ip_usage > 0 and not ip_resource_found:
            self.fatal("Could not find IP v4 resource in Service Template")

    def adjust_pba_resources(self, new_subscription_id):
        if self.adjust_ip and self.adjust_ip > 0:
            self.adjust_resource_in_subscription(new_subscription_id, self.ip_resource, self.adjust_ip)

        if self.with_plesk:
            self.log.info("Updating billing Plesk resource if needed")
            self.increase_plesk_pba(new_subscription_id)

    def check_dns(self):
        self.with_dns = False
        for resource_type in self.legacy_subscription["resource_types"]:
            if resource_type["resclass_name"] == DNS_HOSTING:
                self.with_dns = True

        self.log.info("self.check_dns() is %r" % self.with_dns)


config_file = None
args = None
pleskd_properties = None


# W/A for Python < 3.5 version
def update_migrate_subscription_summary(subscription_id, key, value):
    field = migrate_subscription_summary[str(subscription_id)]
    field[str(key)] = str(value)
    migrate_subscription_summary[str(subscription_id)] = field


def migrate_subscription(subscr_id):
    global fail_counter

    uLogging.init2(loggingUtils.subscription_log_path(subscr_id), True, False)
    migrate_subscription_summary[subscr_id] = {
        "subscrId": subscr_id,
        "newSubscrId": "",
        "targetPlanId": "",
        "targetPlanPeriodId": "",
        "targetServiceTemplateId": "",
        "status": "Not Processed"
    }
    if fail_counter.value >= FAIL_THRESHOLD:
        uLogging.err(
            "[SubId: {}] Migration skipped because FAIL_THRESHOLD has reached (Too many subcriptions failed in this run)".format(
                subscr_id))
        return
    try:
        uLogging.info("[SubId: {}] Starting migration".format(subscr_id))
        app = App(subscr_id, args.configuration_file)
        try:
            app.migrate()
        finally:
            app.db.close()
            app.pba_db.close()
            app.im_db.close()
        update_migrate_subscription_summary(subscr_id, "status", "Ok")
    except Skipped:
        pass
    except Exception:
        uLogging.err("[SubId: {}] {}".format(subscr_id, traceback.format_exc()))
        update_migrate_subscription_summary(subscr_id, "status", "Failed")
        with fail_counter.get_lock():
            fail_counter.value += 1


if __name__ == "__main__":
    loggingUtils.setup_logging(loggingUtils.common_log_path(), DEFAULT_LOGGING_LEVEL)

    parser = argparse.ArgumentParser(description='Migrates legacy VPS subscriptions to OACI subscription. Version: '
                                                 '1.0.23')
    parser.add_argument("--mode",
                        help="Running mode. 'retry-failed-stage' and 'skip-failed-stage' are used to resume failed migration",
                        choices=[Mode.DRY_RUN, Mode.NORMAL_RUN, Mode.RETRY, Mode.SKIP], default=Mode.NORMAL_RUN)

    parser.add_argument("--configuration-file", help="Migration parameters, in JSON format",
                        default="migration_params.json")

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--subscription-id", help="Subscription ID to migrate")
    group.add_argument("--subscription-list", help="File with list of subscription IDs to migrate in a batch")

    args = parser.parse_args()

    if (args.mode == Mode.RETRY or args.mode == Mode.SKIP) and args.subscription_list:
        print " '{}' parameter cannot be used in batch mode. Use --subscription-id to specify single subscription ID".format(
            args.mode)

    if args.mode == Mode.SKIP:
        if not utils.get_user_confirmation(
                "Skip mode: Migration will be run again, the stage that failed at the previous run will be skipped."):
            sys.exit(0)

    elif args.mode == Mode.RETRY:
        if not utils.get_user_confirmation(
                "Retry mode: Migration will be run again, the stage that failed at the previous run will be retried."):
            sys.exit(0)

    subscr_ids_list = []
    if args.subscription_list:
        with open(args.subscription_list) as subscr_ids_list_file:
            for line in subscr_ids_list_file.readlines():
                subscription_id = line.strip()
                if subscription_id not in subscr_ids_list:
                    subscr_ids_list.append(subscription_id)
    else:
        subscr_ids_list.append(args.subscription_id.strip())

    fail_counter = Value('i', 0)
    utils.check_oaci_version(17.13, 422)
    pleskd_properties = utils.load_properties("/usr/local/pem/etc/pleskd.props")
    manager = Manager()
    migrate_subscription_summary = manager.dict()

    try:
        # monitor for pba operations
        pba_lock = manager.Lock()
        thread_pool = Pool(THREAD_POOL_SIZE)
        thread_pool.map(migrate_subscription, subscr_ids_list)
    finally:
        loggingUtils.write_summary(migrate_subscription_summary)
        loggingUtils.create_result_log(migrate_subscription_summary)
        uLogging.info("Done. Log is saved in " + loggingUtils.common_log_path())
