import json
import os
import shutil
import subprocess
import sys
from ConfigParser import ConfigParser
from StringIO import StringIO
from datetime import datetime


def fatal(error, subscription_id=""):
    print "[ERROR] [SubId: {0}] {1}".format(subscription_id, error)
    sys.exit(2)


def debug(msg, subscription_id=""):
    print "[DEBUG] [SubId: {0}] {1}".format(subscription_id, msg)


def mem_size_to_mb(size, subscription_id):
    if not (size.endswith("Mb") or size.endswith("MB")):
        fatal("Unknown units: %s, expected Mb" % (size), subscription_id)

    return int(size[:-2])


class App:
    def __init__(self, ve_id, params, limits, ve_params, subscription_id):
        self.ve_id = ve_id
        self.params = params
        self.limits = limits
        self.ve_params = ve_params
        self.subscription_id = subscription_id

    def stop_ve(self):
        self.old_state = self.ve["State"]
        if self.old_state == "stopped":
            return
        self.shell("prlctl stop %s" % self.ve_id)

    def start_ve(self):
        if self.old_state == "stopped":
            return
        self.shell("prlctl start %s" % self.ve_id)

    # move public ip from routed to bridged interface
    def normalize_net_device(self):
        hw = self.ve["Hardware"]
        for key in hw.keys():
            value = hw[key]
            if "type" in value and value["type"] == "routed":
                if not ("ips" in value):
                    continue

                self.shell("prlctl set %s --device-del %s" % (self.ve_id, key))
                ips = value["ips"].split(" ")
                ip_command = ""
                for ip in ips:
                    if not ip or ip.strip() == "":
                        continue
                    ip_command += " --ipadd %s" % ip
                command = "prlctl set %s --device-add net --network 'Bridged' %s --gw %s" % (
                self.ve_id, ip_command, self.ve_params["gateway"])

                self.shell(command)

    def check_cpu(self):
        cpu_count = self.ve["Hardware"]["cpu"]["cpus"]
        cpu_limit = self.limits["cpu"]

        cpu_exceeds = False
        if (cpu_count == "unlimited"):
            cpu_exceeds = True
        else:
            cpu_count = int(cpu_count)
            if (cpu_count > cpu_limit):
                cpu_exceeds = True

        if cpu_exceeds:
            fatal("SUBSCRIPTION LIMIT EXCEEDED: VPS has %s cpus but subscription limit is %s" % (cpu_count, cpu_limit),
                  self.subscription_id)

    def normalize_cpu(self):
        cpu_count = int(self.ve["Hardware"]["cpu"]["cpus"])
        cpu_limit = self.limits["cpu"]

        if (cpu_count < cpu_limit and self.params["expand_cpu_count_to_subscription_limits"] == "true"):
            self.shell("prlctl set %s --cpus %s --cpuunits 1000 --cpulimit 0" % (self.ve_id, cpu_limit))

    def check_ram(self):
        ram_size = mem_size_to_mb(self.ve["Hardware"]["memory"]["size"], self.subscription_id)
        ram_limit = self.limits["ram"]

        if (ram_size == -1 or ram_size > ram_limit):
            fatal("SUBSCRIPTION LIMIT EXCEEDED: VPS has %s MB of RAM but subscription limit is %s MB" % (
                ram_size, ram_limit), self.subscription_id)

    def is_slm(self):
        debug(self.ve["Home"], self.subscription_id)
        config_file = "%s/ve.conf" % self.ve["Home"]
        debug(config_file, self.subscription_id)
        parser = ConfigParser()
        with open(config_file) as stream:
            stream = StringIO("[top]\n" + stream.read())  # make ConfigParser happy - add a section
            parser.readfp(stream)
            slmmode = parser.get("top", "SLMMODE")
            return slmmode == "\"slm\""

    def normalize_ram(self):
        ram_size = mem_size_to_mb(self.ve["Hardware"]["memory"]["size"], self.subscription_id)
        new_ram_size = ram_size

        if (ram_size < self.limits["ram"] and self.params["expand_ram_to_subscription_limits"] == "true"):
            new_ram_size = self.limits["ram"]

        # need to set --memsize even if it did not change
        self.shell("prlctl set %s --memsize %s" % (self.ve_id, new_ram_size))

    # unset properties that may not be compatible with properties set by OACI
    def normalize_veconf(self):
        cmd = "vzctl unset %s --numproc --numpty --slmmemorylimit --slmmode --vm_overcommit --tcpsndbuf --tcprcvbuf" \
              " --shmpages --quotaugidlimit --quotatime --privvmpages --othersockbuf --numtcpsock --numsiginfo --numpty" \
              " --numproc --numothersock --numiptent --numflock --numfile --lockedpages --kmemsize --iopslimit --ioprio" \
              " --iolimit --dgramrcvbuf --dcachesize --cpulimit --burst_cpulimit --burst_cpu_avg_usage --avnumproc  --save" % self.ve_id

        self.shell(cmd)

    def backup_veconf(self):
        config_dir = self.ve["Home"]
        config_file = "ve.conf"
        suffix = datetime.now().strftime(".%y-%m-%d_%H-%M-%S.%f.backup")

        src_file = os.path.join(config_dir, config_file)
        dist_file = src_file + suffix
        shutil.copy(src_file, dist_file)

    def check_hdd(self):
        hdd_size = mem_size_to_mb(self.ve["Hardware"]["hdd0"]["size"], self.subscription_id)
        hdd_limit = self.limits["hdd"]
        if not (hdd_limit % 1024 == 0):
            fatal("HDD limit in subscription %s MB is not multiple of GB." % (hdd_limit), self.subscription_id)

        if (hdd_size > hdd_limit):
            fatal("SUBSCRIPTION LIMIT EXCEEDED: VPS has %s MB of HDD but subscription limit is %s MB"
                  % (hdd_size, hdd_limit), self.subscription_id)
        elif not (self.params["expand_hdd_to_subscription_limits"] == "true"):
            if not (hdd_size % 1024 == 0):
                fatal("HDD size %s MB is not multiple of GB."
                      "Either resize it manually or or set expand_hdd_to_subscription_limits in migration params"
                      % (hdd_size), self.subscription_id)

    def normalize_hdd(self):
        hdd_size = mem_size_to_mb(self.ve["Hardware"]["hdd0"]["size"], self.subscription_id)
        hdd_limit = self.limits["hdd"]

        if (hdd_size < hdd_limit):
            if (self.params["expand_hdd_to_subscription_limits"] == "true"):
                self.shell("prlctl set %s --device-set hdd0 --size %s" % (self.ve_id, hdd_limit))
            elif not (hdd_size % 1024 == 0):
                fatal("HDD size %s MB is not multiple of GB." % (hdd_size), self.subscription_id)

    def normalize_bw(self):
        bw_limit = self.limits["bandwidth"]
        self.shell("prlctl set %s --rate 1:%s --ratebound yes" % (self.ve_id, bw_limit))

    def normalize(self):

        with open('original--%s.json' % arg_ve_id, 'w') as outfile:
            json.dump(self.ve, outfile, indent=4)

        self.backup_veconf()
        self.normalize_veconf()
        self.normalize_net_device()
        self.normalize_cpu()
        self.normalize_ram()
        self.normalize_hdd()
        self.normalize_bw()

        with open('reconfigured--%s.json' % arg_ve_id, 'w') as outfile:
            json.dump(self.get_ve(), outfile, indent=4)

    def check_limits(self):
        self.ve = self.get_ve()

        self.check_cpu()
        self.check_ram()
        self.check_hdd()

        debug("SUBSCRIPTION LIMITS OK", self.subscription_id)

    def get_ve(self):
        ve_json = self.shell("prlctl list -i %s --json" % self.ve_id)
        ves = json.loads(ve_json)
        return ves[0]

    def shell(self, cmd):
        debug("Executing command: " + cmd, self.subscription_id)
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        r = p.communicate()
        output = r[0]
        error = r[1]
        code = p.returncode

        if code != 0 or error != "":
            raise Exception("Command failed with status %s and error:\n%s" % (code, error))

        return output


if __name__ == "__main__":
    if len(sys.argv) < 5:
        print """Usage: python ve_normalization VE_ID PARAMETERS_FILE SUBSCRIBTION_LIMITS_FILE VE_PARAMS_FILE [-d,--dry-run]
        VE_ID - uuid of VE to be normalized
        PARAMETERS_FILE - migration parameters
        SUBSCRIBTION_LIMITS_FILE - A file path with parameters in JSON format
        VE_PARAMS_FILE - ve parameters file
        SUBSCRIPTION_ID - id of subscription, used for logging, optional.
        """
        sys.exit(1)

    arg_ve_id = sys.argv[1]
    arg_params_json = sys.argv[2]
    arg_limits_json = sys.argv[3]
    arg_ve_params_json = sys.argv[4]
    arg_subscription_id = sys.argv[5] if len(sys.argv) >= 6 else ""

    os.chdir(sys.path[0])

    try:
        with open(arg_params_json) as params_json:
            params = json.load(params_json)

        with open(arg_limits_json) as limits_json:
            limits = json.load(limits_json)

        with open(arg_ve_params_json) as ve_params_json:
            ve_params = json.load(ve_params_json)

    except Exception as e:
        debug(e, arg_subscription_id)
        fatal("Unable to parse parameters JSON", arg_subscription_id)

    app = App(arg_ve_id, params, limits, ve_params, arg_subscription_id)
    app.check_limits()

    if ("--dry-run" in sys.argv) or ("-d" in sys.argv):
        debug("DRY RUN", arg_subscription_id)
    else:
        app.normalize()
