#!/bin/env python

import argparse
import logging
import sys

from poaupdater import uLogging
from utils import DB


logger = None

HOST_NAME_PREFIX = "VPS-MIGR-"


class Host:
    def __init__(self, row):
        self.host_id = int(row[0])
        self.name = row[1]
        self.os = row[2]
        self.ip = row[3] if row[3] else ""
        self.prepared = self.name.startswith(HOST_NAME_PREFIX)


class HostsTable:
    def __init__(self):
        self.s = ""
        self.__add_line("ID", "PREPARED", "NAME", "OS", "IP")

    def add(self, host):
        self.__add_line(str(host.host_id), str(host.prepared), host.name, host.os, host.ip)

    def __add_line(self, *args):
        if self.s != "":
            self.s += "\n"
        self.s += "{:5s}\t{:8s}\t{:40s}\t{:20s}\t{}".format(*args)

    def __repr__(self):
        return self.s


class Application:
    def __init__(self, args):
        self.args = args
        self.db = DB()

    def run(self):
        cmd = self.args.cmd
        if cmd == "list":
            self.run_list()
        elif cmd == "prepare":
            self.run_prepare()
        else:
            raise Exception("Invalid command %s" % cmd)

    def run_list(self):
        logger.info("Listing hosts")
        hosts = self.get_hosts()
        s = self.hosts_to_string(hosts)
        logger.info("Found %s hosts:\n%s" % (len(hosts), s))

    def run_prepare(self):
        host_id = self.args.host_id
        if self.args.host_id is None:
            raise Exception("Host ID is required")
        logger.info("Going to prepare host with id %s" % host_id)
        host = self.get_host(host_id)

        if host is None:
            raise Exception("Unknown host %s" % host_id)

        if host.prepared:
            raise Exception("Host already prepared")

        self.rename_host(host)
        self.remove_ips(host)

    def get_hosts(self):
        return self.db.select("""
select 
h.host_id,h.primary_name,h.platform_desc, string_agg(ips.ip_address,' ')
from hw_nodes as n 
left join hosts as h on h.host_id=n.host_id
left join configured_ips as ips on ips.host_id=n.host_id
group by h.host_id,h.primary_name,h.platform_desc
        """, lambda row: Host(row))

    def hosts_to_string(self, hosts):
        table = HostsTable()
        for host in hosts:
            table.add(host)

        return str(table)

    def get_host(self, host_id):
        hosts = filter(lambda h: h.host_id == host_id, self.get_hosts())
        if len(hosts) == 0:
            return None
        return hosts[0]

    def rename_host(self, host):
        if host.name.startswith(HOST_NAME_PREFIX):
            return
        new_name = HOST_NAME_PREFIX + host.name
        logger.info("Renaming host from '%s' to '%s'" % (host.name, new_name))
        self.db.update("""
        update hosts set primary_name = '%s' where host_id = %s;
        """ % (new_name, host.host_id))

    def remove_ips(self, host):
        if host.ip == "":
            return
        logger.info("Removing ip addresses: %s" % host.ip)
        self.db.update("""
        delete from configured_ips where host_id = %s
        """ % host.host_id)


def set_console_log_level(targetLogger, level):
    for handler in targetLogger.handlers:
        if isinstance(handler, logging.StreamHandler):
            handler.setLevel(level)
            break


def configure_uLogging(targetLogger):
    # w/a uLogging initialization issues
    targetLogger.log_file_stream = None
    uLogging.init_external(targetLogger)


if __name__ == "__main__":
    log_format = '%(asctime)-15s %(levelname)-5s %(message)s'
    logging.basicConfig(format=log_format)
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)

    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--verbose", help="Show debug output", action="store_true")
    parser.add_argument("cmd", help="Command to execute", choices=["list", "prepare"])
    parser.add_argument("--host-id", help="Target host ID. Required for prepare command", type=int)
    parser.add_argument("--log-file", help="File to write logs to", default="/var/log/pa/migration.prepare.nodes.log")

    args = parser.parse_args()

    log_file_handler = logging.FileHandler(args.log_file)
    log_file_handler.setFormatter(logging.Formatter(fmt=log_format))
    log_file_handler.setLevel(logging.DEBUG)

    root.addHandler(log_file_handler)

    if not args.verbose:
        set_console_log_level(root, logging.INFO)

    logger = root
    configure_uLogging(root)

    logger.info("Writing detailed log to file %s" % args.log_file)
    logger.debug("Arguments is %s", args)

    try:
        Application(args).run()
    except Exception as e:
        logger.error(e.message)
        logger.debug(e, exc_info=1)
        sys.exit(1)
