#!/bin/env python

import re
import urllib2
import base64
import socket
import json
import time
import os
import tempfile
import httplib

from datetime import datetime
from poaupdater import uSysDB, uBilling, uLogging
from subprocess import Popen, PIPE, STDOUT

DATA_DIR = "/var/log/pa/vps-migration/"
MIGRATION_STATE_FILE = "migration-state-{}.json"
SUBSCRIPTION_LIMITS_FILE = "new-ve-limits-for-subscr-{}.json"
VPS_PARAMS_FILE_TEMPLATE = "vps-params-for-subscr-{}.json"


def migration_state_path(subscription_id):
    return os.path.join(DATA_DIR, MIGRATION_STATE_FILE.format(subscription_id))


def subscription_limits_path(subscription_id):
    return os.path.join(DATA_DIR, SUBSCRIPTION_LIMITS_FILE.format(subscription_id))


def vps_params_path(subscription_id):
    return os.path.join(DATA_DIR, VPS_PARAMS_FILE_TEMPLATE.format(subscription_id))


class DB:
    def __init__(self, connect=None):
        if connect is None:
            self.db = uSysDB.connect()
            self.custom_connect = False
        else:
            self.db = connect
            self.custom_connect = True

    def select(self, sql, row_mapper):
        cur = self.db.cursor()
        try:
            cur.execute(sql)
            result = []
            for row in cur.fetchall():
                result.append(row_mapper(row))
            return result
        except Exception as e:
            cur.execute("ROLLBACK")
            self.db.commit()
            raise e
        finally:
            cur.close()

    def update(self, sql):
        cur = self.db.cursor()
        try:
            cur.execute(sql)
            self.db.commit()
        except Exception as e:
            cur.execute("ROLLBACK")
            self.db.commit()
            raise e
        finally:
            cur.close()

    def close(self):
        if self.custom_connect:
            self.db.close()
        else:
            uSysDB.close(self.db)


class VpsProperties:
    def __init__(self, row):
        self.vps_name = row[0]
        self.host_id = row[1]
        self.vps_id = row[2]
        self.host_ip_with_leading_zeroes = row[3]
        self.vps_ip_with_leading_zeroes = row[4]
        self.ve_host_id = row[5]


class IpPool:
    def __init__(self, row):
        self.ip_from = row[0]
        self.ip_to = row[1]
        self.gateway = row[2]
        self.pool_name = row[3]


class Task:
    def __init__(self, row):
        self.task_id = row[0]
        self.exec_status = row[1]


class Host:
    def __init__(self, row):
        self.host_id = row[0]


def load_properties(filepath, sep='=', comment_char='#'):
    props = {}
    with open(filepath, "rt") as f:
        for line in f:
            l = line.strip()
            if l and not l.startswith(comment_char):
                key_value = l.split(sep)
                key = key_value[0].strip()
                value = sep.join(key_value[1:]).strip().strip('"')
                props[key] = value
    return props


def get_paci_property(openapi, property_name):
    return openapi.pem.paci.getPACIProperty(property_name=property_name)["value"]


def is_billing_exist():
    try:
        uBilling.PBAConf.get_host_id()
        return True
    except Exception:
        return False


def get_config_parameter(config, parameter_name, default_value):
    if parameter_name in config:
        return config[parameter_name]
    else:
        if not default_value:
            raise Exception("Configuration parameter '%s' is missing" % parameter_name)
        else:
            return default_value


def get_int_config_parameter(config, parameter_name, default_value=None):
    return int(get_config_parameter(config, parameter_name, default_value))


def convert_plan_limit_to_pem_uom(value, uom):
    multiplier_by_uom = {
        "MB": 1024,
        "MB (Additive)": 1024,
        "GB": 1024 * 1024,
        "GB (Additive)": 1024 * 1024,
    }

    if value == -1:
        return value

    multiplier = None
    try:
        multiplier = multiplier_by_uom[uom]
    except:
        pass

    if multiplier:
        return value * multiplier
    else:
        return value


def check_oaci_version(required_version, required_build):
    db = DB()
    try:
        oaci_version = \
            db.select("SELECT name FROM hotfixes WHERE name LIKE 'oaci-%' ORDER BY install_date DESC limit 1",
                      lambda row: row[0])[0]
    finally:
        db.close()
    m = re.search('oaci-([\d\.]*)-(\d*)', oaci_version)
    if not m:
        raise Exception("ABORTING: failed to find installed oaci version")
    major_versoin = m.group(1)
    build = m.group(2)
    if float(major_versoin) < required_version or (
            float(major_versoin) == required_version and int(build) < required_build):
        raise Exception("ABORTING: you need to upgrade OACI to oaci-%s-%s version or higher to continue migration" % (
            required_version, required_build))


def request_to_im(request, im_settings):
    request = "%s/%s" % (im_settings["url"], request)
    uLogging.debug("IM-Request: '{}'".format(request))

    socket.setdefaulttimeout(10)

    http_handler = urllib2.HTTPHandler()
    opener = urllib2.build_opener(http_handler)

    base64string = base64.encodestring('%s:%s' % (im_settings["login"], im_settings["password"]))[:-1]

    req = urllib2.Request(request)
    req.add_header("Authorization", "Basic %s" % base64string)
    req.add_header("Accept", "application/json")

    url_handle = opener.open(req)
    response = json.loads(url_handle.read())
    uLogging.debug("IM-Response: '{}'".format(json.dumps(response, indent=4)))
    return response


def get_request_to_aps(request, aps_api, mn_cert):
    headers = {'Content-Type': 'application/json'}
    uLogging.debug("APS-Request: '{}'".format(request))
    resp = aps_api.GET(request, headers, '', mn_cert)
    uLogging.debug("APS-Response: '{}'".format(resp))
    return resp


def post_request_to_aps(request, headers, aps_api, mn_cert):
    uLogging.debug("APS-Request: POST '{}'".format(request))
    resp = aps_api.POST(request, headers, '', mn_cert)
    uLogging.debug("APS-Response: '{}'".format(resp))
    return resp


def put_request_to_aps(request, data, aps_api, mn_cert):
    headers = {'Content-Type': 'application/json'}
    uLogging.debug("APS-Request: PUT '{}'".format(request))
    resp = aps_api.PUT(request, headers, data, mn_cert)
    uLogging.debug("APS-Response: '{}'".format(resp))
    return resp


def plesk_ka_get_key_info(address, login, password, key):
    request = """<?xml version="1.0" encoding="utf-8"?>
<methodCall>
  <methodName>partner10.getKeyInfo</methodName>
  <params>
    <param>
      <value>
        <struct>
          <member>
            <name>login</name>
            <value>
              <string>%s</string>
            </value>
          </member>
          <member>
            <name>password</name>
            <value>
              <string>%s</string>
            </value>
          </member>
        </struct>
      </value>
    </param>
    <param>
      <value>
        <string>%s</string>
      </value>
    </param>
  </params>
</methodCall>""" % (login, password, key)
    uLogging.debug("xml-rpc request: '{}'".format(request))
    headers = {"Content-type": "text/xml"}
    conn = httplib.HTTPSConnection(address)
    conn.request("POST", "", request, headers)
    response = conn.getresponse()
    data = response.read()
    uLogging.debug("xml-rpc response: '{}'".format(data))
    m = re.search('.*resultCode</name><value><int>(\d+)</int>.*', data)
    if m:
        code = m.group(1)
        uLogging.debug("xml-rpc resultCode: '{}'".format(code))
    else:
        raise Exception("Failed to get resultCode")
    conn.close()
    return code


def wait_for(predicate, delay=5, timeout=120, retry_exceptions=True):
    while True:
        try:
            return predicate()
        except Exception as e:
            if isinstance(e, AssertionError):
                pass
            elif not retry_exceptions:
                raise e

            ex = e
            uLogging.debug("still waiting (%s sec left) due to: %s" % (timeout, str(e)))

        if timeout <= 0:
            raise Exception("wait_for timed out: {}".format(ex))

        time.sleep(delay)
        timeout -= delay


def check_task(db, operation_name, query):
    def check():
        tasks = db.select(query, lambda row: Task(row))
        assert len(tasks) != 0, "%s is not started" % operation_name
        assert tasks[0].exec_status in ["f", "s"], "%s task not finished yet" % operation_name
        if tasks[0].exec_status == "f":
            raise Exception("%s is failed" % operation_name)

    return check


def datetime_to_timestamp(dt):
    return int(round((dt - datetime.fromtimestamp(0)).total_seconds()))


def get_user_confirmation(message):
    message += "\nEnter 'yes' to continue:"
    user_input = raw_input(message)
    return user_input == "yes"


def get_ssh_askpass(userPassword):
    fd, tmpname = tempfile.mkstemp()
    try:
        os.write(fd, "#!/bin/sh\n")
        os.write(fd, "rm -f '%s'\n" % tmpname)
        os.write(fd, "echo -n '%s' ; exit\n" % userPassword)
        os.write(fd, "#EOF\n")
    finally:
        os.close(fd)
    os.chmod(tmpname, 0700)
    return tmpname


def run_remote_cmd(commandText, remoteAddress, userLogin, userPassword):
    tmpname = get_ssh_askpass(userPassword)
    try:
        commandText = commandText.replace("'", """'\\''""")
        remoteCommand = """DISPLAY=:0 SSH_ASKPASS='%s' setsid ssh -T -q -2 -oLogLevel=error \
-oStrictHostKeyChecking=no \
-oCheckHostIP=no \
-oUserKnownHostsFile=/dev/null \
-oPreferredAuthentications=publickey,password,keyboard-interactive \
-l '%s' '%s' '%s'""" % (tmpname, userLogin, remoteAddress, commandText)
        uLogging.debug(remoteCommand)
        pipe = Popen(remoteCommand, stdout=PIPE, stderr=PIPE, shell=True)
        stdout, stderr = pipe.communicate()
        rc = pipe.returncode

        if rc is not None:
            s = int(rc)
        else:
            s = 0
        if stdout:
            uLogging.debug(stdout)
        if stderr:
            uLogging.err(stderr)
        if s != 0:
            raise Exception("Command exited with status %d" % (s,))
    finally:
        if os.path.isfile(tmpname):
            os.unlink(tmpname)
